# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import threading
import inspect
from typing import Type, Callable, NewType, Any
from chipscopy.tcf import protocol
from chipscopy.tcf.channel import Token
from chipscopy.utils.logger import log
from . import Node, CsManager


def get_request_queue(cs_manager, queue_group):
    if not cs_manager.request_queue:
        cs_manager.request_queue = {queue_group: RequestQueue()}
    elif queue_group not in cs_manager.request_queue.keys():
        cs_manager.request_queue[queue_group] = RequestQueue()

    return cs_manager.request_queue[queue_group]


class DoneRequest(object):
    def done_request(self, token, error, result):
        pass

    def __call__(self, token, error, results):
        self.done_request(token, error, results)


class ProgressRequest(object):
    def done_progress(self, token, error, result):
        pass

    def __call__(self, token, error, results):
        self.done_progress(token, error, results)


DoneCallback = NewType("DoneCallback", DoneRequest or Callable[[Any, str or Exception, Any], None])
ProgressCallback = NewType("ProgressCallback", ProgressRequest or Callable[[float], None])
# Token = NewType("Token", token)


class GenericCallback(object):
    def __init__(self, callback):
        self.callback = callback

    def __getattr__(self, attr):
        if attr.startswith("done"):
            return self.callback


def _make_callback(done):
    if inspect.isfunction(done) or inspect.ismethod(done):
        return GenericCallback(done)
    return done


class RequestQueue(object):
    def __init__(self):
        self.deq = collections.deque()

    def add(self, req):
        self.deq.append(req)

    def is_first(self, req):
        return req == self.deq[0]

    def done_request(self, req):
        self.deq.remove(req)
        if len(self.deq) > 0:
            next_req = self.deq[0]
            protocol.invokeLater(next_req._invoke)


class CsRequest(object):
    """
    Handles requests for individual nodes managed by a CsManager, ensuring that a particular request is run
    when the node is available.
    """

    def __init__(
        self,
        cs_manager: CsManager,
        node_id: str = "",
        node_cls: Type[Node] = Node,
        done=None,
        progress=None,
    ):
        self.cs_manager = cs_manager
        self.node_id = node_id
        self.node_cls = node_cls
        self.node = None
        self.done = done
        # If the node supports progress update, this member var should have an instance obj of the
        # CsRequestProgressUpdate class
        self.progress = progress
        self.request_queue = None
        self.run_func = self.default_run_func
        self.run_args = ((), {})
        self.result = None
        self.error = ""
        self.called = False

    def get_request_queue(self, node):
        if not self.request_queue:
            queue_group = node.queue_group
            self.request_queue = get_request_queue(self.cs_manager, queue_group)
            self.request_queue.add(self)
        return self.request_queue

    def __getattr__(self, item):
        self.run_func = item
        return self.__call__

    def __call__(self, *args, **kwargs):

        self.run_args = (args, kwargs)
        self.node = None
        self.request_queue = None
        self.result = None
        self.error = None
        self.called = True

        log.request.info(
            f"Starting request {self} {self.cs_manager} {self.node_id} {self.node_cls} {self.run_args}"
        )

        if not protocol.isDispatchThread():
            protocol.invokeLater(self._invoke)
        else:
            self._invoke()
        return self

    def _invoke(self):
        if not self.cs_manager.is_ready:
            log.request.debug(f"cs_manager not ready {self}")
            protocol.invokeLaterWithDelay(1, self._invoke)
            return

        node = self.cs_manager.get_node(self.node_id)

        if not node:
            self.done_run(None, Exception(f"Could not find context for {self.node_id}"), None)
            return

        request_queue = self.get_request_queue(node)

        if not request_queue.is_first(self):
            log.request.info(f"Queuing {self} {node.queue_group}")
            return

        if not node.is_ready:
            log.request.debug(f"Node not ready {self}")
            protocol.invokeLaterWithDelay(1, self._invoke)
            return

        try:
            node = self.cs_manager.get_node(self.node_id, self.node_cls)
        except Exception as e:
            self.error = e
            self.done_run(self, self.error, self.result)
            return

        if not node:
            self.done_run(
                None,
                Exception(f"Could not find context for {self.node_id} of type {self.node_cls}"),
                None,
            )
            return

        if not node.is_ready:
            log.request.debug(f"Node class switch not ready {self}")
            protocol.invokeLaterWithDelay(1, self._invoke)
            return

        self.node = node
        args, kwargs = self.run_args
        try:
            log.request.info(f"Running {self} {self.run}")
            result = self.run(*args, **kwargs)
            if result is not None:
                self.result = result
        except Exception as e:
            # exception_type = type(e).__name__
            # exception_message = e
            # self.error = f"({exception_type}) {exception_message}"
            self.error = e

        if self.called and self.result is not None or self.error:
            self.done_run(self, self.error, self.result)

    def run(self, *args, **kwargs):
        if isinstance(self.run_func, str):
            self.run_func = getattr(self.node, self.run_func)
        done = kwargs.get("done")
        if not done and args and len(inspect.signature(self.run_func).parameters) == len(args):
            done = args[-1]
            args = args[:-1]
        if done:
            self.done = done
        kwargs["done"] = self.done_run

        # NOTE - Only if the progress_update var has a valid CsRequestProgressUpdate instance obj
        #  as its value, add this to the kwargs. Many existing funcs dont expect to get the
        #  "progress_update" argument and error out.
        if self.progress is not None:
            kwargs["progress_update"] = self.progress_update

        if self.run_func:
            log.request.debug(f"Run func {self.run_func} ({args}, {kwargs}")
            self.run_func(*args, **kwargs)
        else:
            self.error = Exception("No run function set for request")

    def done_run(self, token=None, error=None, result=None):
        log.request.info(f"Done Run {self} {error} {result}")
        self.run_func = self.default_run_func
        if self.request_queue:
            self.request_queue.done_request(self)
        if self.done:
            self.done(self, error, result)
        self.called = False

    def progress_update(self, token=None, error=None, result=None):
        log.request.info(f"Progress Update {self} {result}")
        self.run_func = self.default_run_func
        if self.progress:
            self.progress(self, error, result)

    def default_run_func(self, done=None, progress_update=None):
        self.result = self.cs_manager[self.node_id]


class CsRequestSync(CsRequest):
    def __init__(self, cs_manager: CsManager, node_id: str = "", node_cls: Type[Node] = Node):
        cond = threading.Condition()
        req = self

        class DoneSync(DoneRequest):
            def done_request(self, token, error, result):
                req.error = error
                req.result = result
                with cond:
                    cond.notify()

        super(CsRequestSync, self).__init__(cs_manager, node_id, node_cls, DoneSync())
        self.cond = cond

    def __call__(self, *args, **kwargs):
        timeout = None
        if "timeout" in kwargs:
            timeout = kwargs.pop("timeout")

        # If done is passed as None or not passed at all then wait - meaning block until command completes.
        # This could happen with kwargs (done=arg) or with done passed as the last argument in list.
        # These cases are handled:
        # 1. function_call(blah, blah)                <- No done passed at all in list   : Block
        # 2. function_call(blah, blah, None)          <- Done passed as None as last arg : Block
        # 3. function_call(blah, blah, done=None)     <- Done passed as None with kwarg  : Block
        # 4. function_call(blah, blah, done=callback) <- Done callback as kwarg          : Do not block
        # 5. function_call(blah, blah, callback)      <- Done callback as last arg       : Do not block

        should_wait = True  # Case 1,2,3
        if "done" in kwargs and kwargs["done"] is not None:
            # Case 4
            should_wait = False
        elif self.run_func:
            run_func = self.run_func
            if isinstance(run_func, str):
                run_func = getattr(self.node_cls, run_func)
                if run_func and len(args) == len(inspect.signature(run_func).parameters) - 1:
                    if len(args) > 0 and args[len(args) - 1] is not None:
                        # Case 5
                        should_wait = False

        log.request.info(f"Request sync start {self} timeout {timeout}")
        if not should_wait:
            super(CsRequestSync, self).__call__(*args, **kwargs)
            return self

        with self.cond:
            super(CsRequestSync, self).__call__(*args, **kwargs)
            completed = self.cond.wait(timeout)
        log.request.debug(
            f"Request sync done {self.error} {self.result} completed {completed} {self}"
        )
        if self.error:
            if isinstance(self.error, str):
                self.error = Exception(self.error)
            elif isinstance(self.error, dict):
                self.error = Exception(self.error.get("Format"))
            elif not isinstance(self.error, BaseException):
                self.error = Exception(str(self.error))
            raise self.error
        elif not completed:
            raise Exception("Request timed out.")
        return self.result


class CancelError(Exception):
    pass


DoneFutureCallback = NewType("DoneFutureCallback", Callable[["CsFuture"], None])
ProgressFutureCallback = NewType("ProgressFutureCallback", Callable[["CsFuture"], None])


def null_callback(future):
    pass


class CsFuture(object):
    def __init__(self, done=None, progress=None, final=None):
        self._done_callback = CsFuture._wrap_callback(done)
        self._progress_callback = CsFuture._wrap_callback(progress)
        self.is_done = False
        self._result = None
        self._error = None
        self._progress_status = None
        self.final = final

    def unset(self):
        self.is_done = False
        self._result = None
        self._error = None

    @staticmethod
    def _wrap_callback(done):
        if not done:
            return done

        if inspect.isfunction(done) or inspect.ismethod(done):
            sig = inspect.signature(done)
            if len(sig.parameters) == 1:
                return done

        # done = _make_callback(done)

        def done_oldcallback(future: CsFuture):
            done(future, future._error, future._result)

        return done_oldcallback

    def _sanitize_error(self):
        if isinstance(self._error, str):
            self._error = Exception(self._error)
        elif isinstance(self._error, dict):
            self._error = Exception(self.error.get("Format"))

    def _finalize(self):
        if self.final and not protocol.isDispatchThread():
            final = self.final
            self.final = None
            final(self)

    @property
    def result(self):
        if not self.is_done:
            raise Exception("CsFuture not done")
        if self._error:
            self._sanitize_error()
            raise self._error
        self._finalize()
        return self._result

    @property
    def progress(self):
        if self._error:
            self._sanitize_error()
            raise self._error
        return self._progress_status

    @property
    def error(self):
        if not self.is_done:
            raise Exception("CsFuture not done")
        self._sanitize_error()
        self._finalize()
        return self._error

    def _invoke_done(self):
        self.is_done = True
        if self._done_callback:
            self._done_callback(self)

    def _invoke_progress(self):
        if self._progress_callback:
            self._progress_callback(self)

    def set_result(self, result):
        self._result = result
        self._invoke_done()

    def set_progress(self, progress_status):
        self._progress_status = progress_status
        self._invoke_progress()

    def set_exception(self, error):
        self._error = error
        self._invoke_done()

    def cancel(self):
        if not self._error:
            self.set_exception(CancelError("Request Cancelled"))

    @property
    def handle_old_done(self):
        future = self

        def old_done_handler(token, error, result):
            if error:
                future.set_error(error)
            else:
                future.set_result(result)

        return old_done_handler


class CsFutureSync(CsFuture):
    def __init__(self, done=None, timeout=None, *args, **kwargs):
        super().__init__(done, *args, **kwargs)
        self.cond = threading.Event()
        self.timeout = timeout

    def _invoke_done(self):
        super()._invoke_done()
        if self.cond:
            self.cond.set()

    @property
    def result(self):
        self.cond.wait(self.timeout)
        return super().result

    @property
    def error(self):
        self.cond.wait(self.timeout)
        return super().error

    def run(self, func, *args, **kwargs):
        if protocol.isDispatchThread():
            func(*args, **kwargs)
        else:
            protocol.invokeLater(func, *args, **kwargs)
        if self.cond:
            return self.result
        return self

    def wait(self, timeout: int = None):
        if protocol.isDispatchThread():
            return
        if timeout is not None:
            self.timeout = timeout
        completed = self.cond.wait(self.timeout)
        log.request.debug(
            f"Request sync done {self._error} {self._result} completed {completed} {self}"
        )
        if not completed and not self._error:
            self.cancel()
            raise Exception("Request timed out.")
        else:
            self._finalize()


class CsFutureRequest(CsFuture):
    """
    Handles requests for individual nodes managed by a CsManager, ensuring that a particular request is run
    when the node is available.
    """

    def __init__(
        self, cs_manager: CsManager, node_id: str = "", node_cls: Type[Node] = Node, *args, **kwargs
    ):
        self.node_cls = node_cls
        super().__init__(*args, **kwargs)
        self.cs_manager = cs_manager
        self.node_id = node_id
        self.node = None
        # If the node supports progress update, this member var should have an instance obj of the
        # CsRequestProgressUpdate class
        self.request_queue = None
        self.run_func = None
        self.run_args = ((), {})
        self.called = False

    def get_request_queue(self, node):
        if not self.request_queue:
            queue_group = node.queue_group
            self.request_queue = get_request_queue(self.cs_manager, queue_group)
            self.request_queue.add(self)
        return self.request_queue

    def __getattr__(self, item):
        if hasattr(self.node_cls, item):
            self.run_func = item
            return self.__call__
        return self.__getattribute__(item)

    def unset(self):
        super().unset()
        self.node = None
        self.request_queue = None

    def __call__(self, *args, **kwargs):
        self.unset()
        self.run_args = (args, kwargs)
        self.called = True

        log.request.info(
            f"Starting request {self} {self.cs_manager} {self.node_id} {self.node_cls} {self.run_args}"
        )

        if not protocol.isDispatchThread():
            protocol.invokeLater(self._invoke)
        else:
            self._invoke()
        return self

    def _invoke(self):
        if not self.cs_manager.is_ready:
            log.request.debug(f"cs_manager not ready {self}")
            protocol.invokeLaterWithDelay(1, self._invoke)
            return

        node = self.cs_manager.get_node(self.node_id)

        if not node:
            self.set_exception(Exception(f"Could not find context for {self.node_id}"))
            return

        request_queue = self.get_request_queue(node)

        if not request_queue.is_first(self):
            log.request.info(f"Queuing {self} {node.queue_group}")
            return

        if not node.is_ready:
            log.request.debug(f"Node not ready {self}")
            protocol.invokeLaterWithDelay(1, self._invoke)
            return

        try:
            node = self.cs_manager.get_node(self.node_id, self.node_cls)
        except Exception as e:
            self.set_exception(e)
            return

        if not node:
            self.set_exception(
                Exception(f"Could not find context for {self.node_id} of type {self.node_cls}")
            )
            return

        if not node.is_ready:
            log.request.debug(f"Node class switch not ready {self}")
            protocol.invokeLaterWithDelay(1, self._invoke)
            return

        self.node = node
        args, kwargs = self.run_args
        try:
            log.request.info(f"Running {self} {self.run}")
            node.request = self
            result = self.run(*args, **kwargs)
            if result is not None:
                self._result = result
        except Exception as e:
            # exception_type = type(e).__name__
            # exception_message = e
            # self.error = f"({exception_type}) {exception_message}"
            self._error = e

        if self.called and self._result is not None or self._error:
            self.done_run(self, self._error, self._result)

    def run(self, *args, **kwargs):
        if isinstance(self.run_func, str):
            self.run_func = getattr(self.node, self.run_func)
        done = kwargs.get("done")
        sig = inspect.signature(self.run_func)
        has_done = "done" in sig.parameters.keys()
        if has_done and not done and args and len(sig.parameters) == len(args):
            done = args[-1]
            args = args[:-1]
        if done:
            self._done_callback = CsFuture._wrap_callback(done)

        if has_done:
            kwargs["done"] = self.done_run
        elif done:
            del kwargs["done"]

        # NOTE - Only if the progress_update var has a valid CsRequestProgressUpdate instance obj
        #  as its value, add this to the kwargs. Many existing funcs dont expect to get the
        #  "progress_update" argument and error out.
        if "progress_update" in sig.parameters.keys():
            kwargs["progress_update"] = self.progress_update

        result = None
        if self.run_func:
            log.request.debug(f"Run func {self.run_func} ({args}, {kwargs}")
            result = self.run_func(*args, **kwargs)
            if not self.node.is_ready:
                result = None
        else:
            self._error = Exception("No run function set for request")
        return result

    def _invoke_done(self):
        log.request.info(f"Done Run {self}")
        super()._invoke_done()
        self.node.request = None
        self.run_func = None
        self.node.remove_pending(self)
        if self.request_queue:
            self.request_queue.done_request(self)
        self.called = False

    def done_run(self, token=None, error=None, result=None):
        self._error = error
        self._result = result
        self._invoke_done()

    def progress_update(self, token=None, error=None, result=None):
        if not self._error:
            self._error = error
        self.set_progress(result)

    def default_run_func(self, done=None, progress_update=None):
        self._result = self.cs_manager[self.node_id]


class CsFutureRequestSync(CsFutureRequest):
    def __init__(
        self,
        cs_manager: CsManager,
        node_id: str = "",
        node_cls: Type[Node] = Node,
        timeout: int = None,
        *args,
        **kwargs,
    ):
        super(CsFutureRequestSync, self).__init__(cs_manager, node_id, node_cls, *args, **kwargs)
        self.cond = threading.Event()
        self.timeout = timeout

    def _invoke_done(self):
        super()._invoke_done()
        self.cond.set()

    def wait(self, timeout: int = None):
        if protocol.isDispatchThread():
            return
        if timeout is not None:
            self.timeout = timeout
        completed = self.cond.wait(self.timeout)
        log.request.debug(
            f"Request sync done {self._error} {self._result} completed {completed} {self}"
        )
        if not completed and not self._error:
            self.cancel()
            raise Exception("Request timed out.")
        else:
            self._finalize()

    @property
    def result(self):
        self.wait()
        return super().result

    @property
    def error(self):
        self.wait()
        return super().error

    def __call__(self, *args, **kwargs):
        self.cond.clear()

        if "timeout" in kwargs:
            self.timeout = kwargs.pop("timeout")

        # if done is set then don't wait
        should_wait = True
        if self._done_callback or "done" in kwargs:
            should_wait = False
        elif self.run_func:
            run_func = self.run_func
            if isinstance(run_func, str):
                run_func = getattr(self.node_cls, run_func)
            if run_func:
                sig = inspect.signature(run_func)
                done_index = None
                for i, k in enumerate(sig.parameters.keys()):
                    if k == "done":
                        done_index = i
                        break
                if done_index is not None and len(args) > done_index:
                    should_wait = False

        log.request.info(f"Request sync start {self} timeout {self.timeout}")

        super(CsFutureRequestSync, self).__call__(*args, **kwargs)

        if not should_wait:
            return self

        return self.result



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from chipscopy.tcf.services import svf, DoneHWCommand


class svfProxy(svf.SVFService):
    """TCF SVF service interface."""

    def add_target(self, name: str, done: DoneHWCommand) -> None:
        args = {"name": name}
        return self.send_xicom_command("reqAddTarget", args, done)

    def add_device(self, target_ctx: str, name: str, done: DoneHWCommand) -> str:
        args = {"ctx": target_ctx, "name": name}
        return self.send_xicom_command("reqAddDevice", args, done)

    def remove_target(self, target_ctx: str, done: DoneHWCommand) -> None:
        args = {"ctx": target_ctx}
        return self.send_xicom_command("reqRemoveTarget", args, done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from rich.tree import Tree
from typing import TYPE_CHECKING, ClassVar, Dict, Optional

from chipscopy.api import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.api.containers import QueryList
from chipscopy.api.ibert.aliases import CHILDREN, GT_GROUP_KEY, HANDLE_NAME, IBERT_KEY, TYPE
from chipscopy.api.ibert.gt_group import GTGroup
from chipscopy.api.report import report_hierarchy
from chipscopy.utils import deprecated_api
from typing_extensions import Final, final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.client.ibert_core_client import IBERTCoreClient


@final
class IBERT(DebugCore["IBERTCoreClient"]):
    """
    Main API class to use IBERT (Integrated Bit Error Ratio Tester) debug core.
    """

    def __init__(self, ibert_tcf_node):
        super(IBERT, self).__init__(CoreType.IBERT, ibert_tcf_node)

        self.layout = self.core_tcf_node.get_layout()

        # At the top most level only one key = IBERT name is expected
        name = list(self.layout.keys())[0]
        configuration = self.layout[name]

        # ---------------------------------------------------------------
        # Essential members from SerialObjectBase defined here
        # ---------------------------------------------------------------
        self.name: Final[str] = name
        """Name of the IBERT core"""

        self.type: Final[str] = IBERT_KEY
        """Serial object type"""

        self.handle: Final[str] = configuration[HANDLE_NAME]
        """Handle from cs_server"""

        self.children: QueryList[GTGroup] = QueryList()
        """Children of IBERT"""
        # ---------------------------------------------------------------

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

        if CHILDREN in configuration:
            # Build the child objects
            for child_name, child_configuration in configuration[CHILDREN].items():
                if child_configuration[TYPE] == GT_GROUP_KEY:
                    child_class = GTGroup
                else:
                    continue

                new_child = child_class(
                    name=child_name,
                    parent=self,
                    tcf_node=self.core_tcf_node,
                    configuration=child_configuration,
                )
                self.children.append(new_child)

        self.gt_groups: QueryList[GTGroup] = self.children.filter_by(type=GT_GROUP_KEY)
        """GT Group children of the IBERT core"""

    def __repr__(self) -> str:
        return self.name

    def __rich_tree__(self):
        root = Tree(self.name)

        for child in self.children:
            root.add(child.__rich_tree__())

        return root

    def reset(self):
        """
        Reset all RXs, TXs and PLLs in the GT Groups
        """
        for gt_group in self.gt_groups:
            gt_group.reset()

    @deprecated_api(release="2021.1", replacement="report_hierarchy(<Any serial object>)")
    def print_hierarchy(self):
        """
        Prints hierarchy of IBERTs child objects like the GT Group, GT, PLL etc to stdout.
        The format is similar to the output of the unix command - `tree`
        """
        report_hierarchy(self)



# *****************************************************************************
# * Copyright (c) 2011, 2013 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# AxisILAProxy.py
from typing import Any, Dict, List
from chipscopy.tcf.services import DoneHWCommand, Service
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy


# Service name.
NAME = "AxisILA"
"""
ILA service name.
"""


class AxisILAService(Service):
    def getName(self) -> str:
        """
        Get service name of this service.

        :returns: Service name :const:`NAME`.
        """
        return NAME


class AxisILAProxy(CorePropertyProxy, AxisILAService):
    def __init__(self, channel):
        super().__init__(channel)
        self._listeners = {}

    #
    #  General methods
    #
    def initialize(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("initialize", (node_id,), done)

    def terminate(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("terminate", (node_id,), done)

    #
    # Probe Methods
    #
    def define_probe(
        self, node_id: str, probe_defs: List[Dict[str, Any]], done: DoneHWCommand
    ) -> None:
        return self.send_xicom_command(
            "defineProbe",
            (
                node_id,
                probe_defs,
            ),
            done,
        )

    def define_port_probes(
        self, node_id: str, options: Dict[str, Any], done: DoneHWCommand
    ) -> None:
        return self.send_xicom_command(
            "definePortProbes",
            (
                node_id,
                options,
            ),
            done,
        )

    def undefine_probe(self, node_id: str, probe_names: List[str], done: DoneHWCommand) -> None:
        return self.send_xicom_command(
            "undefineProbe",
            (
                node_id,
                probe_names,
            ),
            done,
        )

    def get_probe(
        self, node_id: str, probe_names: List[str], attrs: List[str], done: DoneHWCommand
    ) -> Dict[str, Dict[str, Any]]:
        return self.send_xicom_command("getProbe", (node_id, probe_names, attrs), done)

    def set_probe(
        self, node_id: str, attrs: Dict[str, Dict[str, Any]], done: DoneHWCommand
    ) -> None:
        return self.send_xicom_command("setProbe", (node_id, attrs), done)

    def reset_probe(
        self,
        node_id: str,
        reset_trigger_values: bool,
        reset_capture_values: bool,
        done: DoneHWCommand,
    ) -> None:
        return self.send_xicom_command(
            "resetProbe", (node_id, reset_trigger_values, reset_capture_values), done
        )

    #
    # Core Communication Methods
    #
    def arm(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("arm", (node_id,), done)

    def get_trigger_registers(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("getTriggerRegisters", (node_id,), done)

    def upload(self, node_id: str, done: DoneHWCommand) -> bool:
        return self.send_xicom_command("upload", (node_id,), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
"""

from typing import Dict, List, Any, NewType
from chipscopy.tcf import channel
from chipscopy.tcf.channel.Command import Command
from chipscopy.tcf.services import Service, DoneHWCommand, Token

SERVICE = "Profiler"


class ProfilerProxy(Service):
    """TCF Profiler service interface."""

    def getName(self):
        """
        Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return SERVICE

    def get_capabilities(self, ctx: str, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Gets the capabilites of a given context.

        :param ctx: Context id of the context
        :param done: Callback with result and any error.
        :return: A dictionary of the capabilities
        """
        return self.send_xicom_command("getCapabilities", (ctx,), done)

    def configure(self, ctx: str, params: Dict[str, Any], done: DoneHWCommand) -> Dict[str, Any]:
        """
        Sets the configuration of a given context.

        :param ctx: Context id of the context
        :param params: Configuration parameters to set
        :param done: Callback with result and any error.
        """
        return self.send_xicom_command("configure", (ctx, params), done)

    def read(self, ctx: str, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Reads data available of a given context.

        :param ctx: Context id of the context
        :param done: Callback with result and any error.
        :return: Data if available
        """
        service = self
        name = "read"
        args = (ctx,)

        class ProfileReadCommand(Command):
            def __init__(self):
                super().__init__(service.channel, service.getName(), name, args)

            def result(self, token, data: bytes):
                result = b""
                error = None
                try:
                    mem = memoryview(data)
                    data_count_start = data.find(b"(")
                    if data_count_start >= 0:
                        data_count_start += 1
                        data_count_end = data.find(b")", data_count_start)
                        if data_count_end >= 0:
                            data_count = int(mem[data_count_start:data_count_end])
                            data_start = data_count_end + 1
                            result = mem[data_start : data_start + data_count]
                except Exception as e:
                    error = e
                self.done(error, result)

            def done(self, error, results):
                if done:
                    done(self.token, error, results)

        return ProfileReadCommand().token



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

noc_nmu_typedef = "noc_nmu"
noc_nsu_typedef = "noc_nsu"
ddrmc_noc_typedef = "ddrmc_noc"
ddrmc_main_typedef = "ddrmc_main"
noc_node_types = [noc_nmu_typedef, noc_nsu_typedef, ddrmc_noc_typedef, ddrmc_main_typedef]
noc_bw_lat_types = [noc_nmu_typedef, noc_nsu_typedef, ddrmc_noc_typedef]



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from chipscopy import dm
from chipscopy.client import connect_xicom
from chipscopy.client.core import get_cs_view, CoreParent
from chipscopy.client.core_property_client import CorePropertyClient
from chipscopy.tcf import protocol
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.proxies.DDRMCProxy import NAME as DDRMC_NAME


class DDRMCClient(CorePropertyClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.type == "ddrmc_main"

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(DDRMC_NAME)

    #
    #  General methods
    #
    def initialize(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.initialize(self.ctx, done_cb))

    def terminate(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.terminate(self.ctx, done_cb))

    def write32(self, write_addr: int, data: int, domain_index: int, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.write32(self.ctx, write_addr, data, domain_index, done_cb)
        return self.add_pending(token)

    def read32(
        self, read_addr: int, read_word_count: int, domain_index: int, done: DoneHWCommand = None
    ):
        service, done_cb = self.make_done(done)
        token = service.read32(self.ctx, read_addr, read_word_count, domain_index, done_cb)
        return self.add_pending(token)

    #
    #  Service specific user interfaces
    #
    def refresh_property(self, names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_property(self.ctx, names, done_cb)
        return self.add_pending(token)

    def refresh_cal_status(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_cal_status(self.ctx, done_cb)
        return self.add_pending(token)

    def refresh_cal_margin(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_cal_margin(self.ctx, done_cb)
        return self.add_pending(token)

    def refresh_health_status(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_health_status(self.ctx, done_cb)
        return self.add_pending(token)

    def refresh_dqs_status(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_dqs_status(self.ctx, done_cb)
        return self.add_pending(token)

    def disable_tracking(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.disable_tracking(self.ctx, done_cb)
        return self.add_pending(token)

    def enable_tracking(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.enable_tracking(self.ctx, done_cb)
        return self.add_pending(token)

    def get_cal_status(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_cal_status(self.ctx, done_cb)
        return self.add_pending(token)

    def get_cal_margin_mode(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_cal_margin_mode(self.ctx, done_cb)
        return self.add_pending(token)

    def get_margin_ps_factors(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_margin_ps_factors(self.ctx, done_cb)
        return self.add_pending(token)

    def is_clock_gated(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.is_clock_gated(self.ctx, done_cb)
        return self.add_pending(token)

    def __check_scan_settings(self, done: DoneHWCommand = None) -> bool:
        rw_mode = self["mgchk_rw_mode"]
        vref_min = self["es_vref_min"]
        vref_max = self["es_vref_max"]

        print("Min Vref is set at: ", vref_min)
        print("Max Vref is set at: ", vref_max)

        if rw_mode:
            if (vref_min > 50) or (vref_max > 50):
                print("ERROR: Cannot set Vref values larger than 50 under Write margin mode.")
                return False

        if vref_min > vref_max:
            print("ERROR: Cannot set minimum Vref value larger than maximum Vref value.")
            return False

        return True

    def run_eye_scan(self, data_checked=False, done: DoneHWCommand = None) -> bool:
        service, done_cb = self.make_done(done)

        # Perform logical data check if has not been done from earlier calls
        if not data_checked:
            data_pass = self.__check_scan_settings()
            if not data_pass:
                token = object()
                protocol.invokeLater(done, token, None, False)
                return token

        token = service.run_eye_scan(self.ctx, done=done_cb, progress=None)
        return self.add_pending(token)

    def get_eye_scan_data(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_eye_scan_data(self.ctx, done_cb)
        return self.add_pending(token)

    def run_eye_scan_async(self, data_checked=False) -> bool:
        assert self.request
        service = self.get_service_proxy()

        # Perform logical data check if has not been done from earlier calls
        if not data_checked:
            data_pass = self.__check_scan_settings()
            if not data_pass:
                result = False
                self.request.set_result(result)

        def update_progress(token, error, results):
            if not error:
                complete_count = results[0]
                total_count = results[1]
                self.request.set_progress(float(complete_count / total_count))

        def check_result(token, error, result):
            # check if canceled
            if not self.request:
                return

            if error:
                self.request.set_exception(error)
            elif result:
                self.request.set_result(result)
            else:
                self.request.set_exception("Internal error occurred while performing 2D eye scan.")

        service.run_eye_scan(self.ctx, done=check_result, progress=update_progress)


def find_ddrmc(cs_url: str, hw_url: str = "", ddrmc_index=0) -> DDRMCClient:

    if hw_url != "":
        server = connect_xicom(cs_url)
        server.connect_remote(hw_url)
        print(f"Connected to cs_server {cs_url} and hw_server {hw_url}")
    else:
        # single argument passed in is intended to be the actual cs_server
        server = cs_url

    print(f"CS Server services: {server.services}")

    # Set up core detection and get list of found cores along with context properties
    cs_view = get_cs_view(server)
    print(f"Searching for DPC in {cs_view}")
    dpc = None
    for node in cs_view.get_children():
        if "DPC" in node.props.get("Name") or "XVC" in node.ctx:
            dpc = cs_view.get_node(node.ctx, CoreParent)
            break

    print()
    print(f"DPC: {dpc}")

    # Set up all available hardened cores
    dpc.setup_cores()

    print()
    print("ChipScope View:")
    cs_view.print_tree(False)

    # Find DDRMC under DPC with a given index
    ddrmc = server.cs_target(parent=dpc, type="ddrmc_main", index=ddrmc_index, cls=DDRMCClient)

    return ddrmc



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from abc import abstractmethod
from typing import List, Tuple

from chipscopy.tcf.services import from_xargs
from .server_info import ServerInfo
from .view_info import ViewInfo
from chipscopy import dm
from chipscopy.dm import chipscope, request


def get_cs_view(server: ServerInfo) -> ViewInfo:
    """
    Gets the Chipscope view for handling debug cores

    :param server: Active Chipscope server for which to get the CS View
    :return: Chipscope View
    """
    if not server or not isinstance(server, ServerInfo):
        raise Exception("Attempted to get view for inactive server info object")
    return server.get_view(chipscope)


class CoreParent(dm.Node):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type

    def setup_cores(self, debug_hub_addrs: List[int] = (), done: request.DoneCallback = None):
        """
        Sets up cores on a given context

        :param debug_hub_addrs: List of address offsets for Debug Hubs
        :param done: DoneRequest callback
        :return: Token of request
        """
        proc = self.manager.channel.getRemoteService("ChipScope")
        done = request._make_callback(done)

        def done_setup(token, error, results):
            self.remove_pending(token)
            if done:
                done.done_request(token, error, results)

        # add pending to indicate that this node is being changed
        return self.add_pending(proc.setup_debug_cores(self.ctx, debug_hub_addrs, done_setup))

    def remove_cores(self, done: request.DoneCallback = None):
        """
        Removes child cores on a given context

        :param done: DoneRequest callback
        :return: Token of request
        """
        proc = self.manager.channel.getRemoteService("ChipScope")
        done = request._make_callback(done)

        def done_remove(token, error, results):
            self.remove_pending(token)
            if done:
                done.done_request(token, error, results)

        # add pending to indicate that this node is being changed
        return self.add_pending(proc.remove_debug_cores(self.ctx, done_remove))


class CoreClient(dm.Node):
    def post_init(self):
        def done_post_init(token, error, results):
            self.remove_pending(token)
            if error:
                # Detect and bypass disabled DDRMC found error to be thrown
                disabled_mc = "DDRMC get_initialize: DDRMC is not enabled"
                if isinstance(error, dict) and (disabled_mc in error["Format"]):
                    pass
                else:
                    raise Exception(error)

        service = self.get_service_proxy()
        if service:
            self.add_pending(service.initialize(self.ctx, done_post_init))

    @abstractmethod
    def get_service_proxy(self):
        pass

    def make_done(self, done, restore_tuple=False) -> Tuple:
        service = self.get_service_proxy()
        done = request._make_callback(done)

        def done_core_command(token, error, results):
            self.remove_pending(token)
            if done:
                # results = from_xargs(results)
                if isinstance(results, list) and len(results) == 1:
                    results = results[0]
                if restore_tuple and isinstance(results, list):
                    results = [None if not item else tuple(item) for item in results]

                done.done_request(token, error, results)

        return service, done_core_command


def progress_callback(func):
    """
    Simple decorator to pass on the progress result and suppress the token and error.
    In case of progress TCF messages, the value of error is same as token.
    """

    def actual_decorator(*args):
        progress_result = args[2]
        func(progress_result)

    return actual_decorator



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from threading import Event
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple

from rich.box import SQUARE as BOX_SQUARE
from rich.table import Table

from chipscopy.api.ibert.aliases import (
    EYE_SCAN_2D_PLOT,
    EYE_SCAN_2D_PLOT_BER_FLOOR_VALUE,
    EYE_SCAN_2D_PLOT_DATA,
    EYE_SCAN_ABORTED,
    EYE_SCAN_DONE,
    EYE_SCAN_ERROR_COUNT,
    EYE_SCAN_HORZ_RANGE,
    EYE_SCAN_IN_PROGRESS,
    EYE_SCAN_NOT_STARTED,
    EYE_SCAN_PRESCALE,
    EYE_SCAN_PROGRESS,
    EYE_SCAN_RAW_DATA,
    EYE_SCAN_SAMPLE_COUNT,
    EYE_SCAN_SCAN_PARAMETERS,
    EYE_SCAN_START_TIME,
    EYE_SCAN_STATUS,
    EYE_SCAN_STOP_TIME,
    EYE_SCAN_TOTAL_NO_OF_DATA_POINTS_EXPECTED,
    EYE_SCAN_TOTAL_NO_OF_DATA_POINTS_READ,
    EYE_SCAN_UT,
    EYE_SCAN_VERT_RANGE,
    MB_ELF_VERSION,
)
from chipscopy.api.ibert.rx import RX
from chipscopy.api.ibert.eye_scan.params import EyeScanParam
from chipscopy.api.ibert.eye_scan.plotter import EyeScanPlot
from chipscopy.utils.printer import printer, PercentProgressBar


if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.dm import Node

common_progress = PercentProgressBar()


@dataclass
class ScanPoint:
    x: int
    y: int
    ber: float
    errors: int
    samples: int


@dataclass(frozen=True)
class Plot2DData:
    scan_points: Dict[Tuple[int, int], ScanPoint]
    """Collection of :py:class:`ScanPoint` instances to represent the 2D eye scan plot"""

    ber_floor_value: float
    """This value is used as the ``ber`` value for any X, Y coordinate whose computed BER was 0"""


@dataclass
class RawData:
    """
    Class for storing raw data from the MicroBlaze. The size of all the lists in this class
    will be the same
    """

    ut: List[int]
    prescale: List[int]
    error_count: List[int]
    sample_count: List[int]
    vertical_range: List[int]
    horizontal_range: List[int]


@dataclass
class ScanData:
    """
    Class for storing raw scan data and 2D plot data
    """

    raw: RawData
    """Raw scan data from the MicroBlaze"""

    all_params: dict
    """All parameters the server used to run the scan in the MicroBlaze"""

    processed: Plot2DData = None
    """2D plot data. This will only be available if the scan completed successfully"""


@dataclass
class EyeScan:
    """
    Class for interacting with eye scans.
    **Please do not** create an instance of this class directly. Please use the factory method
    :py:func:`~chipscopy.api.ibert.create_eye_scans` instead.
    """

    rx: RX
    """:py:class:`RX` object attached to this eye scan"""

    name: str
    """Name of the eye scan"""

    params: Dict[str, EyeScanParam] = field(default_factory=dict)
    """Dictionary containing the py:class:`EyeScanParam` instance for every available param"""

    done_callback: Callable[["EyeScan"], None] = None
    """Callback function called when eye scan has ended"""

    progress_callback: Callable[[float], None] = None
    """Callback function called when eye scan update is received"""

    filter_by: Dict[str, Any] = field(default_factory=dict)

    status: str = EYE_SCAN_NOT_STARTED
    """Status of the eye scan"""

    progress: float = -1.0
    """Progress of the eye scan in %"""

    stop_time: datetime = None
    """Time stamp of when eye scan was stopped in cs_server"""

    start_time: datetime = None
    """Time stamp of when eye scan was started in cs_server"""

    elf_version: str = None
    """ELF version read from the MicroBlaze"""

    data_points_read: int = -1
    """Number of data points scanned by the MicroBlaze"""

    data_points_expected: int = -1
    """Total number of data points expected to be scanned by the MicroBlaze"""

    scan_data: ScanData = None
    """Scan data stored in an instance of the :py:class:`ScanData` class"""

    plot: EyeScanPlot = None
    """`EyeScanPlot` instance for interacting with the plot"""

    _handle_from_cs_server: Optional[str] = None

    _task_id = None

    _scan_done_event: Event = field(default_factory=Event)

    def __repr__(self) -> str:
        return self.name

    def __post_init__(self):
        self.rx.eye_scan = self
        self.rx.eye_scan_names.append(self.name)

        self.filter_by = {"rx": self.rx, "name": self.name}

        if self.rx.link is not None:
            self.filter_by["link"] = self.rx.link

        all_params = self.rx.core_tcf_node.get_eye_scan_parameters(rx_name=self.rx.handle)
        valid_values = all_params["Valid Values"]
        default_values = all_params["Default Value"]
        for param in all_params["Parameter Names"]:
            self.params[param] = EyeScanParam(param)
            if param in valid_values:
                self.params[param].modifiable = True
                self.params[param].valid_values = valid_values[param]
                self.params[param].default_value = default_values[param]

        # Todo - Remove event listener after done with everything
        self.rx.property.endpoint_tcf_node.add_listener(self._scan_update_event_listener)

    def _clear_out_old_data(self):
        self.status = EYE_SCAN_NOT_STARTED
        self.progress = -1.0
        self.stop_time = None
        self.start_time = None
        self.elf_version = None
        self.scan_data = None
        self.data_points_read = -1
        self.data_points_expected = -1

        self._scan_done_event.clear()

    def _get_status(self) -> str:
        if self.status == EYE_SCAN_IN_PROGRESS:
            return PercentProgressBar.Status.IN_PROGRESS.value
        elif self.status == EYE_SCAN_ABORTED:
            return PercentProgressBar.Status.ABORTED.value
        elif self.status == EYE_SCAN_DONE:
            return PercentProgressBar.Status.DONE.value
        else:
            return f"[bold]{self.status}[/]"

    def invalidate(self):
        self.rx.eye_scan = None
        self.rx.eye_scan_names.remove(self.name)

        self.rx = None
        self.name = ""
        self._clear_out_old_data()

    def start(self, *, show_progress_bar: bool = True):
        """
        Send command to cs_server to start the eye scan in the HW.

        Args:
            show_progress_bar: Set to true to show progress bar on stdout

        """
        self._clear_out_old_data()

        scan_params = dict()
        for param in self.params.values():
            if param.value is not None:
                if not param.modifiable:
                    printer(
                        f"Scan parameter '{param.name} is not modifiable! It will be ignored",
                        level="warning",
                    )
                    continue
                scan_params[param.name] = param.value

        self._handle_from_cs_server = self.rx.core_tcf_node.start_eye_scan(
            rx_name=self.rx.handle,
            scan_parameters=scan_params,
        )

        self._task_id = common_progress.add_task(
            description=f"{self.name} progress ",
            status=self._get_status(),
            visible=show_progress_bar,
        )

    def _scan_update_event_listener(self, node: "Node", updated_properties: Set[str]):
        # NOTE - This is called on the TCF event dispatcher thread
        if len(updated_properties) == 0:
            return

        try:
            # TODO - Merge this condition with the one at the top
            if self._handle_from_cs_server not in updated_properties:
                return

            scan_report = node.props[self._handle_from_cs_server]

            # Not expected to change after scan start
            if self.start_time is None and EYE_SCAN_START_TIME in scan_report:
                self.start_time = datetime.strptime(
                    scan_report[EYE_SCAN_START_TIME], "%Y-%m-%d %H:%M:%S.%f"
                )

            # Not expected to change after scan stop
            if self.stop_time is None and EYE_SCAN_STOP_TIME in scan_report:
                self.stop_time = datetime.strptime(
                    scan_report[EYE_SCAN_STOP_TIME], "%Y-%m-%d %H:%M:%S.%f"
                )

            # Not expected to change after scan start
            if self.elf_version is None and MB_ELF_VERSION in scan_report:
                self.elf_version = scan_report[MB_ELF_VERSION]

            # Not expected to change after scan start
            if (
                self.data_points_expected == -1
                and EYE_SCAN_TOTAL_NO_OF_DATA_POINTS_EXPECTED in scan_report
            ):
                self.data_points_expected = int(
                    scan_report[EYE_SCAN_TOTAL_NO_OF_DATA_POINTS_EXPECTED]
                )

            if EYE_SCAN_TOTAL_NO_OF_DATA_POINTS_READ in scan_report:
                self.data_points_read = int(scan_report[EYE_SCAN_TOTAL_NO_OF_DATA_POINTS_READ])

            if EYE_SCAN_SCAN_PARAMETERS in scan_report and EYE_SCAN_RAW_DATA in scan_report:
                new_ut = scan_report[EYE_SCAN_RAW_DATA][EYE_SCAN_UT]
                new_prescale = scan_report[EYE_SCAN_RAW_DATA][EYE_SCAN_PRESCALE]
                new_error_count = scan_report[EYE_SCAN_RAW_DATA][EYE_SCAN_ERROR_COUNT]
                new_sample_count = scan_report[EYE_SCAN_RAW_DATA][EYE_SCAN_SAMPLE_COUNT]
                new_vertical_range = scan_report[EYE_SCAN_RAW_DATA][EYE_SCAN_VERT_RANGE]
                new_horizontal_range = scan_report[EYE_SCAN_RAW_DATA][EYE_SCAN_HORZ_RANGE]

                if self.scan_data is None:
                    self.scan_data = ScanData(
                        raw=RawData(
                            ut=new_ut,
                            prescale=new_prescale,
                            error_count=new_error_count,
                            sample_count=new_sample_count,
                            vertical_range=new_vertical_range,
                            horizontal_range=new_horizontal_range,
                        ),
                        all_params=scan_report[EYE_SCAN_SCAN_PARAMETERS],
                    )

                else:
                    self.scan_data.raw.ut.extend(new_ut)
                    self.scan_data.raw.prescale.extend(new_prescale)
                    self.scan_data.raw.error_count.extend(new_error_count)
                    self.scan_data.raw.sample_count.extend(new_sample_count)
                    self.scan_data.raw.vertical_range.extend(new_vertical_range)
                    self.scan_data.raw.horizontal_range.extend(new_horizontal_range)

            if EYE_SCAN_2D_PLOT in scan_report:
                ber_floor_value = scan_report[EYE_SCAN_2D_PLOT][EYE_SCAN_2D_PLOT_BER_FLOOR_VALUE]

                scan_points: Dict[Tuple[int, int], ScanPoint] = dict()
                for key, data in scan_report[EYE_SCAN_2D_PLOT][EYE_SCAN_2D_PLOT_DATA].items():
                    x, y = [int(coordinate) for coordinate in key.split(", ")]
                    ber = data["BER"]
                    errors = data["Errors"]
                    samples = data["Sample"]

                    data_point = (x, y)

                    if data_point in scan_points:
                        scan_point = scan_points[data_point]

                        # Combine existing BER with new BER
                        scan_point.ber += ber

                        # Combine errors if new error != 0 and old error == 0
                        if errors != 0 and scan_point.errors == 0:
                            scan_point.errors = errors

                    else:
                        scan_points[data_point] = ScanPoint(x, y, ber, errors, samples)

                self.scan_data.processed = Plot2DData(
                    scan_points=scan_points, ber_floor_value=ber_floor_value
                )

            if EYE_SCAN_PROGRESS in scan_report:
                self.progress = round(float(scan_report[EYE_SCAN_PROGRESS]), 2)

            if EYE_SCAN_STATUS in scan_report:
                self.status = scan_report[EYE_SCAN_STATUS]

            # If rich pkg is available send update
            common_progress.update(
                task_id=self._task_id,
                completed=int(self.progress),
                status=self._get_status(),
            )

            if self.status in {EYE_SCAN_ABORTED, EYE_SCAN_DONE}:
                # If scan is done create the plot
                if self.status == EYE_SCAN_DONE:
                    self.plot = EyeScanPlot(self)

                # If user has registered done callback function call it.
                if callable(self.done_callback):
                    try:
                        self.done_callback(self)
                    except Exception as e:
                        printer(
                            f"Unhandled exception during eye scan done callback!\n"
                            f"Exception - {str(e)}",
                            level="warning",
                        )

                self._scan_done_event.set()

            else:
                # If user has registered progress callback function call it.
                if callable(self.progress_callback):
                    try:
                        self.progress_callback(self.progress)
                    except Exception as e:
                        printer(
                            f"Unhandled exception during eye scan progress callback!\n"
                            f"Exception - {str(e)}",
                            level="warning",
                        )

        except Exception as e:
            printer(
                f"Unhandled scan update exception on TCF thread!\nException - {str(e)}",
                level="warning",
            )

    def wait_till_done(self):
        """
        Block current thread execution till the eye scan completes OR it is aborted by cs_server/HW.

        """
        if self._handle_from_cs_server is None:
            raise RuntimeError(f"Please start the scan before waiting for it to finish!")

        self._scan_done_event.wait()
        self._handle_from_cs_server = None

    def stop(self):
        """
        Stop eye scan, that is in-progress in the MicroBlaze

        """
        self.rx.core_tcf_node.terminate_eye_scan(rx_name=self.rx.handle)

    def generate_report(self):
        """
        Generate a report for this eye scan object and send it to the "printer" for printing

        """
        report = Table(title=f"{self.name} report", box=BOX_SQUARE)
        report.add_column("Property", justify="right")
        report.add_column("Value", justify="left")

        report.add_row(f"RX", f"{self.rx.handle}")
        report.add_row(f"Status", self._get_status())
        report.add_row(f"Progress", f"{self.progress}%")
        report.add_row(f"Start time", f"{self.start_time}")
        report.add_row(f"Stop time", f"{self.stop_time}")
        report.add_row(f"MB ELF version", f"{self.elf_version}")
        report.add_row(
            f"#Data points expected",
            f"{self.data_points_expected}",
        )
        report.add_row(f"#Data points read", f"{self.data_points_read}")

        report.add_row("", "")

        scan_params = Table(box=BOX_SQUARE)
        scan_params.add_column("Parameter", justify="right")
        scan_params.add_column("Value", justify="left")

        for param in self.params.values():
            scan_params.add_row(
                f"{param.name}",
                f"{param.default_value if param.value is None else param.value}",
            )

        report.add_row(f"Scan parameters", scan_params)

        printer("\n")
        printer(report)
        printer("\n")



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import ByteString, Dict, Any
from chipscopy.tcf.services import Service, DoneHWCommand, Token

NAME = "ContextXvc"
"""XVC Context service name."""


class ContextXvcProxy(Service):
    """TCF XVC Context service interface."""

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def open(self, host: str, port: str, idcode: int = 0, done: DoneHWCommand = None) -> Token:
        """
        Create an SVF target.

        :param host: Host name or IP of XVC server
        :param port: Port of XVC server
        :param idcode: Optional parameter to force the XVC server to show a specific idcode
        :param done: Callback with result and any error.
        """
        args = {"host": host, "port": port}
        if idcode:
            args["idcode"] = idcode
        return self.send_xicom_command("open", args, done)

    def close(self, ctx_id: str, stop: bool = False, done: DoneHWCommand = None) -> Token:
        """
        Create an SVF target.

        :param ctx_id: CTX id of XVC Context
        :param stop: Stop the XVC server prior to disconnect
        :param done: Callback with result and any error.
        """
        args = {"id": ctx_id}
        if stop:
            args["stop"] = 1
        return self.send_xicom_command("close", args, done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ByteString, Dict, Any
from chipscopy.tcf.services import Service, DoneHWCommand, Token

NAME = "Xicom"


class XicomProxy(Service):
    """TCF Xicom service interface."""

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return NAME

    def __init__(self, channel):
        super(XicomProxy, self).__init__(channel)
        self.listeners = {}

    def get_hw_server_version_info(self, done: DoneHWCommand, **props) -> Token:
        """
        Get's the hw_server version information.

        :param done: Callback when command is complete
        :param props: Property arguments
        :return: Token of request
        """
        return self.send_xicom_command("getHwServerVersionInfo", props, done)

    def config_begin(
        self, ctx: str, props: Dict[str, Any] = None, done: DoneHWCommand = None
    ) -> Token:
        """
        Initiates the configuration process.

        :param ctx: Context ID of the device to configure
        :param props: Configuration properties
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        if not props:
            props = {}
        return self.send_xicom_command("configBegin", (ctx, props), done)

    def config_data(self, ctx: str, data: ByteString, done: DoneHWCommand = None) -> Token:
        """
        Sends chunk of configuration data

        :param ctx: Context ID of the device to configure
        :param data: Data chunk to configure
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        return self.send_xicom_command("configData", (ctx, data), done)

    def config_end(
        self, ctx: str, props: Dict[str, Any] = None, done: DoneHWCommand = None
    ) -> Token:
        """
        Ends the configuration process.

        :param ctx: Context ID of the device to configure
        :param props: Configuration properties
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        if not props:
            props = {}
        return self.send_xicom_command("configEnd", (ctx, props), done)

    def config_reset(
        self, ctx: str, props: Dict[str, Any] = None, done: DoneHWCommand = None
    ) -> Token:
        """
        Resets the device configuration.

        :param ctx: Context ID of the device to configure
        :param props: Configuration reset properties
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        if not props:
            props = {}
        return self.send_xicom_command("configReset", (ctx, props), done)

    def secure_debug(self, ctx: str, data: ByteString, done: DoneHWCommand = None) -> Token:
        """
        Enables secure debug using authentication data.

        :param ctx: Context ID of the device
        :param data: Authentication data
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        return self.send_xicom_command("secureDebug", (ctx, data), done)

    def config_in(self, ctx, start_pos, total_bytes, data, done):
        data_count = len(data)
        return self.send_command("configIn", (ctx, start_pos, total_bytes, data_count, data), done)

    def config_start(self, ctx, done):
        return self.send_command("configStart", (ctx,), done)

    def jtag_reg_list(self, ctx: str or int, done: DoneHWCommand = None) -> Token:
        """
        Retrieves list of available JTAG register context IDs

        :param ctx: Context ID of device or idcode of a jtag device
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        return self.send_xicom_command("jtagRegList", (ctx,), done)

    def jtag_reg_def(self, reg_ctx: str, done: DoneHWCommand = None) -> Token:
        """
        Retrieves the definition of a given JTAG register context ID

        :param reg_ctx: JTAG register context ID
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        return self.send_xicom_command("jtagRegDef", (reg_ctx,), done)

    def jtag_reg_get(self, ctx: str, reg_ctx: str, done: DoneHWCommand = None) -> Token:
        """
        Retrieves the current value for a given register

        :param ctx: Context ID of device
        :param reg_ctx: JTAG register context ID
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        return self.send_xicom_command("jtagRegGet", (ctx, reg_ctx), done)

    def readback_config(
        self, ctx: str, props: Dict[str, Any] = None, done: DoneHWCommand = None
    ) -> Token:
        """
        Reads back configuration through the SBI that has been previously set up with another PDI.

        :param ctx: Context ID of the device
        :param props: Property arguments
        +-------------------+----------------------+------------------------------------------+
        | Name              | Type                 | Description                              |
        +===================+======================+==========================================+
        | size              | |int|                | Byte size of the readback                |
        +-------------------+----------------------+------------------------------------------+
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        if not props:
            props = {}
        return self.send_xicom_command("readbackConfig", (ctx, props), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from collections import deque
from itertools import islice
from typing import List, Set, Type, Callable, ClassVar
from chipscopy import dm
from chipscopy.dm import request
from chipscopy.tcf import protocol
from chipscopy.utils.logger import log


class TargetFilter(object):
    def __init__(
        self,
        view_info: "ViewInfo",
        parent: dm.Node or str = None,
        node_cls: ClassVar[dm.Node] = dm.Node,
        **params,
    ):
        self.view_info = view_info
        self._filter_params = params
        self.node_cls = node_cls
        self.parent = parent
        it = view_info.get_children(parent) if parent is not None else view_info.get_all()
        self._filtered_targets = filter(self._filter_props, it)

    def __iter__(self):
        return self

    def __next__(self):
        return self.view_info.get_sync_node(next(self._filtered_targets).ctx, self.node_cls)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.get(index=item)
        elif isinstance(item, str):
            return self.get(ctx=item)
        raise ValueError

    def _filter_props(self, node):
        return (
            node.ctx
            and self.node_cls.is_compatible(node)
            and self._filter_params.items() <= node._props.items()
        )

    def filter(self, parent: dm.Node or str = None, node_cls: ClassVar[dm.Node] = None, **params):
        if parent is None:
            parent = self.parent
        if not node_cls:
            node_cls = self.node_cls
        return TargetFilter(self.view_info, parent, node_cls, **params)

    def get(
        self,
        ctx: str = "",
        parent: dm.Node or str = None,
        node_cls: ClassVar[dm.Node] = None,
        index: int = None,
        **params,
    ):
        if ctx:
            return self.view_info.get_node(ctx)
        if parent is None:
            parent = self.parent
        if not node_cls:
            node_cls = self.node_cls
        if not params:
            params = self._filter_params
        tf = TargetFilter(self.view_info, parent, node_cls, **params)
        if index is not None:
            return next(islice(tf, index, None))
        try:
            node = next(tf)
        except StopIteration:
            # raise StopIteration(f"No nodes for filter {tf}")
            return None
        try:
            _ = next(tf)
            raise Exception(f"More than one result for filter {tf}")
        except StopIteration:
            pass
        return node

    def first(self):
        return self.get(index=0)


class SyncNode(dm.Node):
    def __init__(self, node_id: str = "", parent_ctx: str = "", manager: "ViewInfo" = None):
        super(SyncNode, self).__init__(node_id, parent_ctx, manager)
        self._changed_props = set()
        self._has_changed = False

    def __getattr__(self, attr):
        self.update_changed_props()
        value = super(SyncNode, self).__getattr__(attr)
        if not value and hasattr(self.node_cls, attr):
            return getattr(
                request.CsRequestSync(self.manager.cs_manager, self.ctx, self.node_cls), attr
            )
        return value

    def __getitem__(self, item):
        self.update_changed_props()
        return super(SyncNode, self).__getitem__(item)

    def __setitem__(self, key, value):
        self._props[key] = value

    def __str__(self):
        self.update_changed_props()
        cls = self.node_cls if type(self) == SyncNode else type(self)
        return "{}({}, {} properties)".format(cls.__name__, self.ctx, len(self._props))

    def future(self, *args, **kwargs) -> request.CsFutureRequest:
        return request.CsFutureRequestSync(
            self.manager.cs_manager, self.ctx, self.node_cls, *args, **kwargs
        )

    @property
    def props(self):
        self.update_changed_props()
        return self._props

    def add_changed_props(self, _changed_props):
        self._changed_props.update(_changed_props)
        self._has_changed = True

    def update_changed_props(self):
        sync_node = self
        self.manager.run_events()

        class PropsRequest(request.CsRequestSync):
            def run(self):
                for prop_name in sync_node._changed_props:
                    sync_node[prop_name] = copy.copy(self.node[prop_name])
                sync_node._changed_props.clear()
                sync_node._has_changed = False
                return True

        if self._has_changed:
            req = PropsRequest(self.manager.cs_manager, self.ctx)
            req()


class ViewInfo(dm.CsManager):
    def __init__(self, cs_manager: dm.CsManager):
        super(ViewInfo, self).__init__(cs_manager.name + ".observer", cs_manager.channel)
        self.cs_manager = cs_manager
        self.default_node_cls = SyncNode
        self.next_target_id = 0
        self.ctx_target_map = {}
        self.target_ctx_map = {}
        self.event_queue = deque()
        self.running_events = False
        mi = self

        class NodeListener(dm.NodeListener):
            def nodes_added(self, nodes: List[dm.Node]):
                if log.is_domain_enabled("view_info", "INFO"):
                    names = [node.ctx for node in nodes]
                    log.view_info.info(f"{mi.name}: Adding nodes {names}")
                for node in nodes:
                    props = copy.copy(node._props)
                    mi.queue_event(
                        mi.add_sync_node, node.ctx, node.parent_ctx, node.__class__, props
                    )
                mi.queue_event(mi._node_added)

            def nodes_removed(self, nodes: List[dm.Node]):
                if log.is_domain_enabled("view_info", "INFO"):
                    names = [node.ctx for node in nodes]
                    log.view_info.info(f"{mi.name}: Removing nodes {names}")
                for node in nodes:
                    mi.queue_event(mi.remove_node, node.ctx)
                mi.queue_event(mi._node_removed)  # notify view node listeners

            def node_changed(self, node: dm.Node, updated_keys: Set[str]):
                log.view_info.info(f"{mi.name}: Changing node {node.ctx}: {updated_keys}")
                mi.queue_event(mi.node_changed, node.ctx, copy.copy(updated_keys))

        cs_manager.add_node_listener(NodeListener())

        class ManagerListener(dm.CsManagerListener):
            def manager_added(self, cs_manager):
                if cs_manager == mi.cs_manager:
                    mi.invalid = False

            def manager_removed(self, cs_manager):
                if cs_manager == mi.cs_manager:
                    mi.invalidate()

        ManagerListener.add_listner(ManagerListener())

        class GetNodesRequest(request.CsRequestSync):
            def add_children(self, parent=None):
                # start by creating the parent<->child links in the non-tcf thread view class
                # if there isn't a parent (top level nodes) then we use the manager instance
                # before chipscopy #28 was fixed the TCF thread manager was correctly linking children
                # but the view running on the main thread class was not getting the correct linking
                # but only when the server was caching state before the chipscopy connected
                # https://gitenterprise.xilinx.com/chipscope/chipscopy/issues/28
                if not parent:
                    parent = self.cs_manager
                elif isinstance(parent, str):
                    parent = self.cs_manager[parent]
                mi.queue_event(mi.set_children, parent.ctx, parent.children)
                for child in self.cs_manager.get_children(parent):
                    props = copy.copy(child.props)
                    mi.queue_event(
                        mi.add_sync_node, child.ctx, child.parent_ctx, child.__class__, props
                    )
                    self.add_children(child)

            def run(self):
                self.add_children()
                return True

        req = GetNodesRequest(cs_manager)
        req()
        self.run_events()

    def queue_event(self, event, *args, **kwargs):
        log.view_info.debug(f"{self.name}: queue_event {event}")
        self.event_queue.appendleft((event, args, kwargs))

    def run_events(self):
        if self.running_events:
            return
        try:
            self.running_events = True
            while True:
                event, args, kwargs = self.event_queue.pop()
                log.view_info.debug(f"{self.name}: run_event {event}")
                event(*args, **kwargs)
        except IndexError:
            pass
        finally:
            self.running_events = False

    def node_changed(self, ctx, updated_keys):
        try:
            sync_node = self[ctx]
            sync_node.add_changed_props(updated_keys)
            self._node_changed(sync_node)
        except KeyError:
            pass

    def _node_changed(self, node: SyncNode):
        if node._changed_props:
            for listener in self._node_listeners:
                listener.node_changed(node, node._changed_props)

    def get_sync_node(self, ctx: str, sync_node_cls: ClassVar[SyncNode]):
        return super().get_node(ctx, sync_node_cls)

    def add_sync_node(self, ctx, parent_ctx, cls, props) -> dm.Node:
        sync_node = self.add_node(ctx, parent_ctx)
        if not isinstance(sync_node, SyncNode):
            sync_node = self.upgrade_node(sync_node, SyncNode)
        sync_node["node_cls"] = cls
        sync_node._props.update(props)

        if sync_node.ctx not in self.ctx_target_map.keys():
            self.ctx_target_map[sync_node.ctx] = self.next_target_id
            self.target_ctx_map[self.next_target_id] = sync_node.ctx
            self.next_target_id += 1

        sync_node["target_id"] = self.ctx_target_map[sync_node.ctx]
        return sync_node

    def notify_node_added(self, node: SyncNode):
        if self._node_listeners:
            if node.ctx not in self._added_nodes:
                self._added_nodes.append(node.ctx)
        else:
            node.clear_update()

    def notify_node_removed(self, node):
        if self._node_listeners:
            if node not in self._removed_nodes:
                self._removed_nodes.append(node)

    def notify_node_changed(self, node):
        if not self._node_listeners:
            node.clear_update()

    def get_node(
        self, ctx: str, cls: Type[dm.Node] = dm.Node, done: request.DoneCallback = None
    ) -> dm.Node or None:
        self.run_events()
        node = self._nodes.get(ctx)

        if not node:
            if done:
                protocol.invokeLater(done, None, None, None)
            return None

        if not done and issubclass(node.node_cls, cls):
            return node

        if done:
            req = request.CsRequest(self.cs_manager, ctx, cls, done)
        else:
            req = request.CsRequestSync(self.cs_manager, ctx, cls)

        def _get_node(done):
            node_cls = dm.Node
            if req.node:
                node_cls = req.node.__class__
            req.node["node_cls"] = node_cls
            protocol.invokeLater(done, node, None, req.node)
            return node

        req.run_func = _get_node
        req()

        return node

    def get_children(self, parent: str or dm.Node = ""):
        self.run_events()
        return super(ViewInfo, self).get_children(parent)

    def get_all(self):
        self.run_events()
        return super(ViewInfo, self).get_all()

    def __getitem__(self, item):
        self.run_events()
        return super(ViewInfo, self).__getitem__(item)

    @property
    def size(self):
        self.run_events()
        return super(ViewInfo, self).size

    def find_nodes(self, parent: str or dm.Node = "", cls: Type[dm.Node] = dm.Node):
        """
        Finds a node given a parent node.  If no parent node then searches the root.
        :param parent: Parent node or handle of parent
        :param cls: Desired class
        :return: Nodes of desired class if found.  Empty list otherwise.
        """
        # nodes = [self.get_node(node.ctx, cls) for node in self.get_children(parent) if cls.is_compatible(node)]
        # return nodes
        for node in self.get_children(parent):
            if cls.is_compatible(node):
                yield self.get_node(node.ctx, cls)

    def print_tree(self, print_props: bool = False, printer: Callable[[str], None] = print):
        """
        Prints node tree
        :param print_props: Flag to include properties for each node
        :param printer: Optional argument to change where to print.  (E.g. log.<logger name>.debug)
        """
        self.run_events()

        def print_targets(parent=None, level=1):
            for child in self.get_children(parent):
                name = child.Name if child.Name else child.ctx
                if print_props:
                    printer("{}{} {} ({})".format("\t" * level, child.target_id, name, child.props))
                else:
                    printer("{}{} {}".format("\t" * level, child.target_id, name))
                print_targets(child, level + 1)

        print_targets()



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

from .. import locator
from ... import protocol, peer, channel
from ...channel.Command import Command


class Peer(peer.TransientPeer):
    def __init__(self, parent, attrs):
        super(Peer, self).__init__(attrs)
        self.parent = parent

    def openChannel(self):
        assert protocol.isDispatchThread()
        c = self.parent.openChannel()
        c.redirect(self.getID())
        return c


class ChannelEventListener(channel.EventListener):
    def __init__(self, proxy):
        self.proxy = proxy
        self.channel = proxy.channel

    def event(self, name, data):
        if self.proxy.peers is None:
            return  # peers not synchronized yet

        try:
            args = channel.fromJSONSequence(data)
            if name == "peerAdded":
                assert len(args) == 1
                _peer = Peer(self.channel.getRemotePeer(), args[0])
                if self.proxy.peers.get(_peer.getID()):
                    protocol.log("Invalid peerAdded event", Exception())
                    return
                self.proxy.peers[_peer.getID()] = _peer
                for l in self.proxy.listeners:
                    try:
                        l.peerAdded(_peer)
                    except Exception as x:
                        protocol.log("Unhandled exception in Locator listener",
                                     x)
            elif name == "peerChanged":
                assert len(args) == 1
                m = args[0]
                if not m:
                    raise Exception("Locator service: invalid peerChanged " +
                                    "event - no peer ID")
                _peer = self.proxy.peers.get(m.get(peer.ATTR_ID))
                if not _peer:
                    return
                self.proxy.peers[_peer.getID()] = _peer
                for l in self.proxy.listeners:
                    try:
                        l.peerChanged(_peer)
                    except Exception as x:
                        protocol.log("Unhandled exception in Locator listener",
                                     x)
            elif name == "peerRemoved":
                assert len(args) == 1
                peerID = args[0]
                _peer = self.proxy.peers.get(peerID)
                if not _peer:
                    return
                del self.proxy.peers[peerID]
                for l in self.proxy.listeners:
                    try:
                        l.peerRemoved(peerID)
                    except Exception as x:
                        protocol.log("Unhandled exception in Locator listener",
                                     x)
            elif name == "peerHeartBeat":
                assert len(args) == 1
                peerID = args[0]
                _peer = self.proxy.peers.get(peerID)
                if not _peer:
                    return
                for l in self.proxy.listeners:
                    try:
                        l.peerHeartBeat(peerID)
                    except Exception as x:
                        protocol.log("Unhandled exception in Locator listener",
                                     x)
            else:
                raise IOError("Locator service: unknown event: " + name)
        except Exception as x:
            import sys
            x.tb = sys.exc_info()[2]
            self.channel.terminate(x)


class LocatorProxy(locator.LocatorService):
    def __init__(self, channel):
        self.channel = channel
        self.peers = None  # not yet synchronized
        self.listeners = []
        self.get_peers_done = False
        self.event_listener = ChannelEventListener(self)
        channel.addEventListener(self, self.event_listener)

    def getAgentID(self, done):
        done = self._makeCallback(done)
        service = self

        class GetAgentIDCommand(Command):
            def __init__(self):
                super(GetAgentIDCommand, self).__init__(service.channel,
                                                        service, "getAgentID",
                                                        None)

            def done(self, error, args):
                agentID = None
                if not error:
                    assert len(args) == 2
                    error = self.toError(args[0])
                    agentID = args[1]
                done.doneGetAgentID(self.token, error, agentID)
        return GetAgentIDCommand().token

    def getPeers(self):
        assert protocol.isDispatchThread()
        return self.peers  # None if not synchronized yet

    def redirect(self, _peer, done):
        done = self._makeCallback(done)
        service = self

        class RedirectCommand(Command):
            def __init__(self):
                super(RedirectCommand, self).__init__(service.channel, service,
                                                      "redirect", [_peer])

            def done(self, error, args):
                if not error:
                    assert len(args) == 1
                    error = self.toError(args[0])

                    if not error and service.peers:
                        # The redirect was a success. Invalidate all the peers
                        # already detected.

                        for peerId in list(service.peers.keys()):
                            del service.peers[peerId]
                            for l in service.listeners:
                                try:
                                    l.peerRemoved(peerId)
                                except Exception as x:
                                    protocol.log("Unhandled exception in "
                                                 "Locator listener", x)

                        assert len(service.peers) == 0
                        service.peers = None
                        service.get_peers_done = False

                done.doneRedirect(self.token, error)
        return RedirectCommand().token

    def sync(self, done):
        done = self._makeCallback(done)
        service = self

        class SyncCommand(Command):
            def __init__(self):
                super(SyncCommand, self).__init__(service.channel, service,
                                                  "sync", None)

            def done(self, error, args):
                if error:
                    service.channel.terminate(error)
                done.doneSync(self.token)
        return SyncCommand().token

    def addListener(self, listener):
        self.listeners.append(listener)
        if not self.get_peers_done:
            service = self

            class GetPeersCommand(Command):
                def __init__(self):
                    super(GetPeersCommand, self).__init__(service.channel,
                                                          service, "getPeers",
                                                          None)

                def done(self, error, args):
                    if not error:
                        assert len(args) == 2
                        error = self.toError(args[0])
                    if error:
                        protocol.log("Locator error", error)
                        return
                    c = args[1]
                    service.peers = {}  # peers list synchronized
                    if c:
                        for m in c:
                            peerID = m.get(peer.ATTR_ID)
                            if service.peers.get(peerID):
                                continue
                            _peer = Peer(service.channel.getRemotePeer(), m)
                            service.peers[peerID] = _peer
                            for l in service.listeners:
                                try:
                                    l.peerAdded(_peer)
                                except Exception as x:
                                    protocol.log("Unhandled exception in " +
                                                 "Locator listener", x)
            GetPeersCommand()
            self.get_peers_done = True

    def removeListener(self, listener):
        self.listeners.remove(listener)

    def get_remote_peers(self, done):
        return self.send_command("getPeers", None, done)

    def get_remote_channels(self, done):
        return self.send_command("getChannels", None, done)

    def connect_remote_peer(self, peer_attrib, done):
        return self.send_command("connectPeer", (peer_attrib,), done)

    def disconnect_remote_peer(self, peer_attrib, done):
        return self.send_command("disconnectPeer", (peer_attrib,), done)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from chipscopy import dm
from chipscopy.client.core import CoreClient
from chipscopy.dm.harden.noc_perfmon.traffic_classes import PerfMonTrafficQOSClass
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.proxies.NoCPerfMonProxy import NOC_PERFMON_SERVICE_NAME


class NoCPerfMonCoreClient(CoreClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.type == "npi_nir"

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(NOC_PERFMON_SERVICE_NAME)

    def refresh_property_group(self, groups: List[str], done: DoneHWCommand = None) -> List[str]:
        """

        Args:
            groups (List[str]): list of property group names. For NoC PerfMon the groups are ``control``, ``status``,
                and ``perfmon``.
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            Property name/value pairs from the requested group(s).

        """
        service, done_cb = self.make_done(done)
        token = service.refresh_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    def initialize(self, done: DoneHWCommand = None):
        """
        This method collects the clocking information for client use later in the configuration of this service. It
        also refreshes the client's view of specific properties for the root node *ONLY*.

        This call must be invoked before clocking information is requested from the device

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            No data is returned directly from this method.

        """
        service, done_cb = self.make_done(done)
        return self.add_pending(service.initialize(self.ctx, done_cb))

    #
    # discover the NoC elements in the device
    #
    def discover_noc_elements(self, done: DoneHWCommand = None) -> List[str]:
        """
        This method dynamically scans the NoC to determine the active nodes in the programmed design. If no
        nodes are programmed or the NPI domain is inaccessible this will return an empty list.

        This list will serve as the set from which elements may be chosen for runtime performance analysis.

        Unrouted, disabled NoC elements cannot be configured for measurement. It may be assumed their counters would
        always produce zero count values.

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            Returns a list of active NMU, NSU/DDRMC_NOC, and NPS elements. Unrouted, disabled elements are not returned
            and cannot be monitored.

        """
        service, done_cb = self.make_done(done)
        return self.add_pending(service.discover_noc_elements(self.ctx, done_cb))

    def configure_monitors(
        self,
        monitors: {str},
        sampling_intervals: {dict},
        traffic_class: PerfMonTrafficQOSClass,
        sample_count: int,
        done: DoneHWCommand = None,
        extended_monitor_config: dict = None,
    ):

        """
        This method configures and enables the performance monitoring on the parameter set of elements. Upon completion
        of the configuration, data will start to be emitted asynchronously from ChipScope Server back to the requesting
        client. The client must implement an Event Listener to properly receive the data.

        Each element will be configured to monitor both read and write traffic independently. QOS classes of a
        particular direction of traffic will be applied to the monitor responsible for that type of traffic
        ``read|write``.

        Args:
            monitors (set[str]):
                Set of NoC elements to enable to stream performance data.
                For Example -
                ``('NOC_NMU128_X0Y0', 'DDRMC_X0Y0', 'NOC_NPS5555_X11Y2')``
            sampling_intervals (dict):
                The requested ``period_ms`` for sampling. The client must ensure this is a supported period.
                **NOTE**: this structure needs to contain information for NPI, NoC and each DDRMC controller clk
                NPI, NoC are sampling periods (ms) (floats) and DDRMC controller clk domain is a tap (int)
                -- for reference see the XSA class in the utilities to understand how the ctrl clk works.
            traffic_class (PerfMonTrafficQOSClass):
                The traffic classes bitwise-or'd that shall be monitored.
            sample_count (int):
                *OPTIONAL* The number of intervals or windows to capture. 1 - will configure supporting hardware for
                single shot. 2-int_max will configure all hardware for continuous mode and will stop streaming data when
                the requested number of samples have been sent to the client. -1 is the default and means the streaming
                will run indefinitely.
            filter_spec (dict[int]):
                Will move! hold off on using this for now.
                *OPTIONAL* Dictionary defining the AXI filtering to use. If this parameter is not supplied then no
                filtering will occur and all transactions will be counted by the samplers. No validation is done on
                these fields, they will be masked and shifted, as they are received. For convenience any fields not
                specified will retain the default value of zero, except max fields (AxSIZE, AxQoS, AxLEN) which have the
                largest supported values). The supported fields include

                ``AxPROT[3]``, ``AxBURST[2]``, ``AxSIZE_MIN[3]``, ``AxLEN_MIN[8]``, ``AxID[16]``
                ``TDEST[10]`` (AXIS only), ``DST_ID[12]``, ``AxLOCK[1]``, ``AxQoS_MIN[4]``, ``AxCACHE[4]``,
                ``AxSIZE_MAX[3]``, ``AxQoS_MAX[4]``, ``AxLEN_MAX[8]``

            done:
                Optional command callback that will be invoked when the response is received.

        Returns:
            No data is returned from this call.

        """
        params = {
            "node_id": self.ctx,
            "monitors": monitors,
            "intervals": sampling_intervals,
            "traffic_class": traffic_class,
            "sample_count": sample_count,
            "extended_monitor_config": extended_monitor_config,
        }
        service, done_cb = self.make_done(done)
        return self.add_pending(service.configure_monitors(params, done_cb))

    def get_supported_sampling_periods(
        self,
        ref_clk_freq_mhz: float = None,
        pl_alt_ref_clk_freq_mhz: float = None,
        ddrmc_freq_mhz: dict = {},
        done: DoneHWCommand = None,
    ) -> dict:
        """

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            Dictionary with two top level keys ``NoC`` and ``NPI`` whose values are lists of supported sampling
            periods measured in milliseconds (float).
            :param ddrmc_freq_mhz:
            :param done:
            :param pl_alt_ref_clk_freq_mhz:
            :param ref_clk_freq_mhz:


        """
        params = {"node_id": self.ctx}
        if ref_clk_freq_mhz is not None:
            params.update({"ref_clk_freq_mhz": ref_clk_freq_mhz})
        if pl_alt_ref_clk_freq_mhz is not None:
            params.update({"pl_alt_ref_clk_freq_mhz": pl_alt_ref_clk_freq_mhz})
        params.update(ddrmc_freq_mhz)
        service, done_cb = self.make_done(done)
        return self.add_pending(service.get_supported_sampling_periods(params, done_cb))

    def get_clk_info(self, done: DoneHWCommand = None) -> dict:
        """
        Reports NPI and NoC clock information from the running design. These are parameters set by the designer in
        Vivado.

        Args:
            done: Optional command callback that will be invoked when the response is received.

        Returns:
            A dictionary with two top level keys ``NoC`` and ``NPI`` whose values are dictionaries that
            detail the ``freq_mhz``, frequency in MHz (float) and the ``period_ms``, period in milliseconds (float).

        """
        service, done_cb = self.make_done(done)
        return self.add_pending(service.get_clk_info(self.ctx, done_cb))

    def enumerate_noc_elements(self, nodes, done: DoneHWCommand = None) -> List[str]:
        params = {
            "node_id": self.ctx,
            "scan_nodes": nodes,
        }
        service, done_cb = self.make_done(done)
        return self.add_pending(service.enumerate_noc_elements(params, done_cb))



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""
TCF - Target Communication Framework
"""

import types

from . import compat, protocol, peer, channel
from .util import task

__all__ = ('connect', 'peers')


def connect(params, wait=True):
    """Connect to peer. Argument is a string of the form
    <transport>:<host>:<port>, e.g. "TCP:127.0.0.1:1534".
    """
    if isinstance(params, compat.strings):
        params = _parse_params(params)
    elif isinstance(params, dict):
        raise TypeError("Expected string or dict")
    p = peer.TransientPeer(params)
    if wait:
        c = task.Task(_openChannel, p).get()
    else:
        c = protocol.invokeAndWait(p.openChannel)
    return c


def peers():
    "Return list of discovered remote peers"
    locator = protocol.getLocator()
    if locator:
        return protocol.invokeAndWait(locator.getPeers)


def _openChannel(p, done=None):
    assert protocol.isDispatchThread()
    c = p.openChannel()
    if done is None:
        return

    class ChannelListener(channel.ChannelListener):
        def onChannelOpened(self):
            c.removeChannelListener(self)
            done(None, c)

        def onChannelClosed(self, error):
            done(error, None)
    c.addChannelListener(ChannelListener())


def process_param_str(params, default_port="3121"):
    parts = params.split(":", 3)
    if len(parts) < 3:
        parts.insert(0, "TCP")
    if len(parts) < 3:
        parts.append(default_port)
    return ":".join(parts)


def _parse_params(paramStr):
    args = process_param_str(paramStr).split(":")
    if len(args) != 3:
        raise ValueError("Expected format: <transport>:<host>:<port>")
    transp, host, port = args
    return {
        peer.ATTR_ID: paramStr,
        peer.ATTR_IP_HOST: host,
        peer.ATTR_IP_PORT: port,
        peer.ATTR_TRANSPORT_NAME: transp
    }



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import ByteString, List, Iterable, Union, NewType, Callable, Dict
from enum import Flag, auto
from chipscopy import xvc
from chipscopy.utils import bytes_from_words, words_from_bytes, word_align, WORD_SIZE


class RW(Flag):
    READ = auto()
    WRITE = auto()
    RW = READ | WRITE


class UnwritableException(Exception):
    def __init__(self, reg: "XVCRegister"):
        self.reg = reg

    def __str__(self):
        return f"Unable to write register {self.reg}"


class XVCRegister(object):
    def __init__(self, addr: int, length: int = WORD_SIZE, rw: RW = RW.RW, value: ByteString = b""):
        """

        :param addr:
        :param length:
        :param rw:
        :param value:
        """
        self.addr = addr
        self.length = length
        self.rw = rw
        self._value = bytearray(self.length)
        self.set_value(0, value)

    def valid_addr(self, addr: int):
        return self.addr >= addr < self.addr + self.length

    def read_reg(self) -> ByteString:
        """
        Designed for custom reg override
        :return: Data for reg
        """
        return b""

    def write_reg(self):
        """
        Designed for custom reg override
        The byte value is stored in self._value.  Can also use self.get_words and self.get_word
        to simplify reading value
        """
        pass

    def get_reg(self, addr: int, length: int) -> ByteString:
        length = min(length, self.length)
        if not self.valid_addr(addr) or not self.rw & RW.READ:
            return bytes(length)
        offset = addr - self.addr
        data = self.read_reg()
        if data:
            return data[offset : offset + length]
        if not self._value:
            return bytes(length)
        return self._value[offset : offset + length]

    def set_reg(self, addr: int, data: ByteString) -> int:
        if not self.valid_addr(addr) or not self.rw & RW.WRITE:
            raise UnwritableException(self)
        length = self.set_value(addr - self.addr, data)
        self.write_reg()
        return length

    def set_value(self, offset: int, data: ByteString) -> int:
        if not self._value:
            return self.length
        length = min(len(data), self.length - offset)
        for i in range(length):
            self._value[offset + i] = data[i]
        return length

    def get_words(self, index: int = 0, num_words: int = 1) -> List[int]:
        offset = index * WORD_SIZE
        length = num_words * WORD_SIZE
        if not offset < self.length and offset + length > self.length:
            raise IndexError
        words = words_from_bytes(self._value[offset : offset + length])
        return words

    def set_words(self, words: Iterable[int], index: int = 0):
        offset = index * WORD_SIZE
        data = bytes_from_words(words)
        length = len(data)
        if not offset < self.length and offset + length > self.length:
            raise IndexError
        self.set_value(offset, data)

    def get_word(self, index: int = 0):
        return self.get_words(index)[0]

    def set_word(self, word: int, index: int = 0):
        self.set_words((word,), index)


class XVCRegisterHandler(xvc.XVCHandler):
    def __init__(self, regs: Iterable[XVCRegister], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.registers = [reg for reg in sorted(regs, key=lambda x: x.addr)]

    def get_mem(self, flags: int, addr: int, num_bytes: int):
        addr = word_align(addr)
        data = bytearray()
        try:
            i = iter(self.registers)
            reg = next(i)
            while num_bytes > 0:
                # get next register
                while reg.addr + reg.length <= addr:
                    reg = next(i)
                if reg.addr <= addr < reg.addr + reg.length:
                    reg_data = reg.get_reg(addr, num_bytes)
                    length = len(reg_data)
                    data += reg_data
                else:
                    # padd up to reg.addr
                    length = min(reg.addr - addr, num_bytes)
                    pad_data = bytes(length)
                    data += pad_data
                addr += length
                num_bytes -= length
        except StopIteration:
            # no more regs, pad remaining data
            data += bytes(num_bytes)

        return data

    def set_mem(self, flags: int, addr: int, data: ByteString):
        addr = word_align(addr)
        offset = 0
        length = len(data)
        i = iter(self.registers)
        try:
            reg = next(i)
            while offset < length:
                # get next register
                while reg.addr + reg.length <= addr:
                    reg = next(i)
                if reg.addr <= addr < reg.addr + reg.length:
                    num_bytes = min(length - offset, reg.length)
                    num_bytes = reg.set_reg(addr, data[offset : offset + num_bytes])
                else:
                    # padd up to reg.addr
                    num_bytes = min(reg.addr - addr, length - offset)
                offset += num_bytes
                addr += num_bytes
        except StopIteration:
            pass

    def add_reg(self, new_registers: Union[XVCRegister, Iterable[XVCRegister]]):
        if not isinstance(new_registers, list):
            new_registers = [new_registers]
        self.registers.extend(new_registers)
        self.registers = [reg for reg in sorted(self.registers, key=lambda x: x.addr)]


class XVCCore(XVCRegister):
    def __init__(self, core_type=0xFF, value_byte_len=0x40, drv_ver=0x1):
        super().__init__(0, value_byte_len)
        self.core_type = core_type
        self.drv_ver = drv_ver
        self.core_minor_ver = 1
        self.core_major_ver = 19
        self.tool_minor_ver = 1
        self.tool_major_ver = 19
        self.uuid = bytearray(range(16))

    def handle_core_info(self):
        self.set_value(
            0,
            struct.pack(
                "<xxxx2H4B16s",
                self.core_type,
                self.drv_ver,
                self.core_minor_ver,
                self.core_major_ver,
                self.tool_minor_ver,
                self.tool_major_ver,
                self.uuid,
            ),
        )

    def write_reg(self):
        # handle packet and write back to value
        words = words_from_bytes(self._value)
        if words[0] == 0xC0000003:
            self.handle_core_info()


class XVCDebugHub(XVCRegisterHandler):
    def __init__(
        self,
        cores: Iterable[XVCCore] = (),
        addr: int = 0,
        length: int = 0x1FFFFF,
        axi_data_width=0x0,
        *args,
        **kwargs,
    ):
        hub = self

        class CoreInfoPacket(XVCRegister):
            """
            Reg[7:0]   - This register contains the number of debug core connected to the debug hub over a pair of
                         AXI4-Stream interfaces.
            Reg[15:8]  - Points to the current debug core being written to.
            Reg[23:16] - Points to the current debug core being read from.
            Reg[31:24] - Points to the data width of the AXI4-MM interface as configured.
                         0x0 - 512bits;
                         0x1 - 256bits;
                         0x2 - 128bits;
                         0x4 - 64bits;
                         0x8 - 32bits
            """

            def read_reg(self):
                num_cores = len(hub.cores)
                cur_write_core = 0
                cur_read_core = 0
                return bytes_from_words(
                    (
                        ((hub.axi_data_width & 0xF) << 24)
                        | ((cur_read_core & 0xF) << 16)
                        | ((cur_write_core & 0xF) << 8)
                        | (num_cores & 0xF),
                    )
                )

        class CoreTypeDrv(XVCRegister):
            def read_reg(self):
                word = (hub.drv_ver & 0xFF) << 16 | (hub.core_type & 0xFF)
                return bytes_from_words((word,))

        class HubVersion(XVCRegister):
            def read_reg(self):
                word = (hub.tool_major_ver & 0xF) << 24
                word |= (hub.tool_minor_ver & 0xF) << 16
                word |= (hub.core_major_ver & 0xF) << 8
                word |= hub.core_minor_ver & 0xF
                return bytes_from_words((word,))

        class UUID(XVCRegister):
            def read_reg(self):
                return hub.uuid

        # set up core addresses
        for i, core in enumerate(cores):
            core.addr = 0x1000 * i

        self.cores = [core for core in cores]

        self.axi_data_width = axi_data_width
        self.core_type = 0x1
        self.drv_ver = 0x1
        self.core_minor_ver = 0
        self.core_major_ver = 1
        self.tool_minor_ver = 0
        self.tool_major_ver = 1
        self.uuid = b"0123456789ABCDEF"

        regs = [core for core in cores]
        regs += [
            CoreInfoPacket(0x100000, rw=RW.READ),
            CoreTypeDrv(0x100004, rw=RW.READ),
            HubVersion(0x100008, rw=RW.READ),
            UUID(0x100040, 16, rw=RW.READ),
            XVCRegister(0x1000C0),  # control_reg
        ]

        super(XVCDebugHub, self).__init__(regs, addr, length, *args, **kwargs)


# A handler handles a particular packet opcode and app. specific index.
PacketHandler = NewType("PacketHandler", Callable[[XVCCore, List[int]], None])


class XVCAxisCore(XVCCore):
    def __init__(self, core_type=0x1, value_byte_len=0x90, drv_ver=0x1):
        XVCCore.__init__(self, core_type, value_byte_len=value_byte_len, drv_ver=drv_ver)
        # key is "write packet header" word, as a number.
        self.packet_handlers: Dict[int, PacketHandler] = {}
        self.set_packet_handler(XVCAxisCore.core_info_handler, op_code=0)

    def core_info_handler(self, _: [int]) -> None:
        self.handle_core_info()

    @staticmethod
    def make_write_header_word(op_code: int, index: int) -> int:
        #
        #  http://confluence.xilinx.com/display/XSW/Debug+IP+proposal+for+transaction+exceeding+burst+length
        #
        #  Bits 31:30 - read/write operation. "Read without any specific start address" binary value "11".
        #                                     "Read with start address" binary value "10"
        #  Bits 29:20 - application specific usage (for ILA/VIO, e.g. MU/TC/PROBE_OUT index)
        #  Bits 19:14 - opcode
        #  Bits 13:0  - (sent_byte_count - 1), byte_length includes header word.
        #
        header_word = 0xC0000003 + (op_code << 14) + (index << 20)
        return header_word

    def set_packet_handler(
        self, packet_handler: PacketHandler, op_code: int, index: int = 0, header_word=None
    ):
        hdr = (
            header_word
            if header_word is not None
            else XVCAxisCore.make_write_header_word(op_code, index)
        )
        self.packet_handlers[hdr] = packet_handler

    def get_packet_handler(self, packet_hdr: int) -> PacketHandler:
        return self.packet_handlers.get(packet_hdr, None)

    def write_reg(self):
        # handle packet and write back to value
        words = words_from_bytes(self._value)
        handler = self.get_packet_handler(words[0])
        if not handler:
            raise Exception(f"XVCAxisCore bad packet header: 0x{words[0]:08X}")
        handler(self, words)

    def set_return_packet(self, packet_words: [int]) -> None:
        self.set_value(0, bytes_from_words(packet_words))


class XVCVersalRegisterHandler(XVCRegisterHandler):
    def __init__(self, regs: Iterable[XVCRegister], *args, **kwargs):
        registers = {x.addr: x for x in regs}
        registers[0xF1260114] = XVCRegister(0xF1260114, value=bytes_from_words((0x02000000,)))
        registers[0xF1110884] = XVCRegister(0xF1110884, value=bytes_from_words((0x1,)))
        super().__init__(registers.values(), *args, **kwargs)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict, Any, Iterable
from chipscopy import dm
from chipscopy.dm import request
from chipscopy.utils import words_from_bytes, bytes_from_words, WORD_SIZE
from chipscopy.tcf.services import memory as service
from chipscopy.proxies.ContextXvcProxy import NAME as ContextXvcService
from chipscopy.proxies.DebugCoreProxy import DebugCoreService
from chipscopy.proxies.ProfilerProxy import SERVICE as ProfilerService


class Profiler(object):
    def __init__(self, ctx: str, name: str, capabilities: Dict[str, Any]):
        self.ctx = ctx
        self.name = name
        self.capabilities = capabilities
        self.configuration = {}
        self.data = b""

    def __getitem__(self, item):
        return self.capabilities[item]

    def __setitem__(self, key, value):
        self.configuration[key] = value

    def __str__(self):
        return f"{self.capabilities}"

    def __repr__(self):
        return self.__str__()

    def configure(self, params: Dict[str, Any]):
        self.configuration = params


class MemoryNode(dm.Node):
    """
    Implements memory based transactions for compatible nodes.
    """

    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.AccessTypes and "data" in node.AccessTypes

    def mrd(self, address: int, size: int = 1, done: request.DoneRequest = None):
        """
        Reads requested words from memory context
        :param address: Byte address of starting read
        :param size: Number of words to read
        :param done: Callback when operation complete
        :return: Token of request
        """
        assert done
        done = request._make_callback(done)
        mem = self.mem
        if not mem:
            raise Exception("Could not find memory context for node {}".format(self.ctx))

        address = int(address / WORD_SIZE) * WORD_SIZE
        size *= WORD_SIZE
        mode = service.MODE_CONTINUEONERROR | service.MODE_VERIFY | service.MODE_BYPASS_ADDR_CHECK
        buf = bytearray(size)

        class DoneMemory(service.DoneMemory):
            def doneMemory(self, token, error):
                error = str(error) if error else ""
                if done:
                    done.done_request(token, error, words_from_bytes(buf))

        return mem.get(address, WORD_SIZE, buf, 0, size, mode, DoneMemory())

    def mwr(
        self, address: int, words: List[int] or int, size: int = 1, done: request.DoneRequest = None
    ):
        """
        Writes given words to memory context.
        :param address: Starting byte address of write
        :param words: List of words to write
        :param size: Number of words to write
        :param done: Callback when complete
        :return: Token of request
        """
        assert done
        done = request._make_callback(done)
        mem = self.mem
        if not mem:
            raise Exception("Could not find memory context for node {}".format(self.ctx))

        if isinstance(words, int):
            words = (words,)

        size = max(size, len(words))

        buf = bytes_from_words(words, size)
        address = int(address / WORD_SIZE) * WORD_SIZE
        size *= WORD_SIZE
        mode = service.MODE_CONTINUEONERROR | service.MODE_VERIFY | service.MODE_BYPASS_ADDR_CHECK

        class DoneMemory(service.DoneMemory):
            def doneMemory(self, token, error):
                error = str(error) if error else ""
                if done:
                    done.done_request(token, error, {})

        return mem.set(address, WORD_SIZE, buf, 0, size, mode, DoneMemory())

    def write_bytes(
        self,
        address: int,
        data: bytearray,
        offset: int = 0,
        byte_size: int = None,
        done: request.DoneRequest = None,
    ):
        """
        Writes given words to memory context.
        :param address: Write starting byte address
        :param data: bytearray to write
        :param offset: Start index in *data* to start writing from.
        :param byte_size: Number of bytes to write. Default is size of the *data* buffer.
        :param done: Callback when complete
        :return: Token of request
        """
        assert done
        done = request._make_callback(done)
        mem = self.mem
        if not mem:
            raise Exception(f"Could not find memory context for node {self.ctx}")
        if byte_size is None:
            byte_size = len(data)
        if byte_size > len(data) - offset:
            raise Exception(
                "Request to write more data than provided, to memory context {self.ctx}."
            )

        address = (address // WORD_SIZE) * WORD_SIZE
        mode = service.MODE_CONTINUEONERROR | service.MODE_VERIFY | service.MODE_BYPASS_ADDR_CHECK

        class DoneMemory(service.DoneMemory):
            def doneMemory(self, token, error):
                error = str(error) if error else ""
                if done:
                    done.done_request(token, error, {})

        return mem.set(address, WORD_SIZE, data, offset, byte_size, mode, DoneMemory())

    def get_profilers(self) -> Dict[str, Any] or request.CsFuture:
        assert request
        service = self.manager.channel.getRemoteService(ProfilerService)

        def done_get_capabilities(token, error, results):
            # check if canceled
            if not self.request:
                return

            # check for error or positive result
            if error:
                self.request.set_exception(error)
            else:
                profilers = {key: Profiler(self.ctx, key, value) for key, value in results.items()}
                self.request.set_result(profilers)

        service.get_capabilities(self.ctx, done_get_capabilities)

    def run_profilers(self, profilers: Iterable):
        assert request
        service = self.manager.channel.getRemoteService(ProfilerService)
        params = {p.name: p.configuration for p in profilers}

        def done_configure(token, error, results):
            # check if canceled
            if not self.request:
                return

            # check for error or positive result
            if error:
                self.request.set_exception(error)
            else:
                self.request.set_result(results)

        service.configure(self.ctx, params, done_configure)

    def stop_profilers(self, profilers: Iterable[Profiler]):
        assert request
        service = self.manager.channel.getRemoteService(ProfilerService)
        params = {p.name: {} for p in profilers}

        def done_configure(token, error, results):
            # check if canceled
            if not self.request:
                return

            # check for error or positive result
            if error:
                self.request.set_exception(error)
            else:
                self.request.set_result(results)

        service.configure(self.ctx, params, done_configure)

    def read_profilers(self, profilers: Iterable[Profiler]):
        assert request
        service = self.manager.channel.getRemoteService(ProfilerService)

        def done_read(token, error, results):
            # check if canceled
            if not self.request:
                return

            # check for error or positive result
            if error:
                self.request.set_exception(error)
            else:
                for p in profilers:
                    p.data = results
                self.request.set_result(None)

        service.read(self.ctx, done_read)


def open_xvc(cs_manager: dm.CsManager, host: str, port: str, done: request.DoneCallback):
    """
    Connects to XVC server and opens the XVC Memory Context
    :param cs_manager: Memory Context View Manager
    :param host: Host name or IP Address of XVC Server
    :param port: Port of XVC Server
    :param done: Callback when complete
    :return: Token of request
    """
    proc = cs_manager.channel.getRemoteService(ContextXvcService)
    done = request._make_callback(done)

    # link the ContextXvcService callback directly to this function's callback
    # since we don't need to do any post processing
    return proc.open(host, port, done.done_request)


class XvcNode(MemoryNode):
    """
    Implements functionality to open and close an XVC Memory Context.
    It inherits from MemoryNode to also support mrd and mwr.
    """

    @staticmethod
    def is_compatible(node):
        return node.Name and "XVC" in node.Name and super(XvcNode, XvcNode).is_compatible(node)

    def close(self, done: request.DoneCallback):
        """
        Closes an XVC Memory Context and disconnects from the XVC server
        :param done: Callback when complete
        :return: Token of request
        """
        proc = self.manager.channel.getRemoteService(ContextXvcService)
        done = request._make_callback(done)
        node = self

        def done_cmd(token, error, args):
            node.remove_pending(token)  # done state change
            done.done_request(token, error, args)

        # add pending to indicate that this node is being changed
        return self.add_pending(proc.close(self.ctx, done=done_cmd))
